import * as React from 'react';
import classNames from 'classnames';
import { createAction, createReducer } from '@reduxjs/toolkit';
import { createContext, useContextSelector } from 'use-context-selector';
import FormulaParser, { DepParser, FormulaError } from 'fast-formula-parser';

/*! *****************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */
/* global Reflect, Promise */

var extendStatics = function(d, b) {
    extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (Object.prototype.hasOwnProperty.call(b, p)) d[p] = b[p]; };
    return extendStatics(d, b);
};

function __extends(d, b) {
    if (typeof b !== "function" && b !== null)
        throw new TypeError("Class extends value " + String(b) + " is not a constructor or null");
    extendStatics(d, b);
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
}

var __assign = function() {
    __assign = Object.assign || function __assign(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p)) t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};

function __rest(s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
}

function __generator(thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
}

function __values(o) {
    var s = typeof Symbol === "function" && Symbol.iterator, m = s && o[s], i = 0;
    if (m) return m.call(o);
    if (o && typeof o.length === "number") return {
        next: function () {
            if (o && i >= o.length) o = void 0;
            return { value: o && o[i++], done: !o };
        }
    };
    throw new TypeError(s ? "Object is not iterable." : "Symbol.iterator is not defined.");
}

function __read(o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
}

function __spreadArray(to, from) {
    for (var i = 0, il = from.length, j = to.length; i < il; i++, j++)
        to[j] = from[i];
    return to;
}

var setData = createAction("SET_DATA", function (data) { return ({ payload: { data: data } }); });
var selectEntireRow = createAction("SELECT_ENTIRE_ROW", function (row, extend) { return ({ payload: { row: row, extend: extend } }); });
var selectEntireColumn = createAction("SELECT_ENTIRE_COLUMN", function (column, extend) { return ({
    payload: { column: column, extend: extend }
}); });
var selectEntireTable = createAction("SELECT_ENTIRE_TABLE");
var select = createAction("SELECT", function (point) { return ({ payload: { point: point } }); });
var activate = createAction("ACTIVATE", function (point) { return ({ payload: { point: point } }); });
var setCellData = createAction("SET_CELL_DATA", function (active, data) { return ({
    payload: { active: active, data: data }
}); });
var setCellDimensions = createAction("SET_CELL_DIMENSIONS", function (point, dimensions) { return ({
    payload: { point: point, dimensions: dimensions }
}); });
var copy = createAction("COPY");
var cut = createAction("CUT");
var paste = createAction("PASTE", function (data) { return ({ payload: { data: data } }); });
var edit$1 = createAction("EDIT");
var view$1 = createAction("VIEW");
var clear$1 = createAction("CLEAR");
var blur$1 = createAction("BLUR");
var keyPress = createAction("KEY_PRESS", function (event) { return ({ payload: { event: event } }); });
var keyDown = createAction("KEY_DOWN", function (event) { return ({ payload: { event: event } }); });
var dragStart = createAction("DRAG_START");
var dragEnd = createAction("DRAG_END");
var commit$1 = createAction("COMMIT", function (changes) { return ({ payload: { changes: changes } }); });

/**
 * Creates an empty matrix with given rows and columns
 * @param rows - integer, the amount of rows the matrix should have
 * @param columns - integer, the amount of columns the matrix should have
 * @returns an empty matrix with given rows and columns
 */
function createEmpty(rows, columns) {
    var matrix = Array(rows);
    for (var i = 0; i < rows; i++) {
        matrix[i] = Array(columns);
    }
    return matrix;
}
/** Gets the value at row and column of matrix. */
function get$1(point, matrix) {
    var columns = matrix[point.row];
    if (columns === undefined) {
        return undefined;
    }
    return columns[point.column];
}
/** Creates a slice of matrix from startPoint up to, but not including, endPoint. */
function slice(startPoint, endPoint, matrix) {
    var sliced = [];
    var columns = endPoint.column - startPoint.column;
    for (var row = startPoint.row; row <= endPoint.row; row++) {
        var slicedRow = row - startPoint.row;
        sliced[slicedRow] = sliced[slicedRow] || Array(columns);
        for (var column = startPoint.column; column <= endPoint.column; column++) {
            sliced[slicedRow][column - startPoint.column] = get$1({ row: row, column: column }, matrix);
        }
    }
    return sliced;
}
/** Sets the value at row and column of matrix. If a row doesn't exist, it's created. */
function set$2(point, value, matrix) {
    var nextMatrix = __spreadArray([], __read(matrix));
    // Synchronize first row length
    var firstRow = matrix[0];
    var nextFirstRow = firstRow ? __spreadArray([], __read(firstRow)) : [];
    if (nextFirstRow.length - 1 < point.column) {
        nextFirstRow[point.column] = undefined;
        nextMatrix[0] = nextFirstRow;
    }
    var nextRow = matrix[point.row] ? __spreadArray([], __read(matrix[point.row])) : [];
    nextRow[point.column] = value;
    nextMatrix[point.row] = nextRow;
    return nextMatrix;
}
/** Like Matrix.set() but mutates the matrix */
function mutableSet(point, value, matrix) {
    var firstRow = matrix[0];
    if (!firstRow) {
        firstRow = [];
        matrix[0] = firstRow;
    }
    if (!(point.row in matrix)) {
        matrix[point.row] = [];
    }
    // Synchronize first row length
    if (!(point.column in firstRow)) {
        firstRow[point.column] = undefined;
    }
    matrix[point.row][point.column] = value;
}
/** Removes the coordinate of matrix */
function unset$1(point, matrix) {
    if (!has$3(point, matrix)) {
        return matrix;
    }
    var nextMatrix = __spreadArray([], __read(matrix));
    var nextRow = __spreadArray([], __read(matrix[point.row]));
    // Avoid deleting to preserve first row length
    nextRow[point.column] = undefined;
    nextMatrix[point.row] = nextRow;
    return nextMatrix;
}
/** Creates an array of values by running each element in collection thru iteratee. */
function map$1(func, matrix) {
    var e_1, _a;
    var newMatrix = [];
    try {
        for (var _b = __values(entries$2(matrix)), _c = _b.next(); !_c.done; _c = _b.next()) {
            var _d = __read(_c.value, 2), point = _d[0], value = _d[1];
            mutableSet(point, func(value, point), newMatrix);
        }
    }
    catch (e_1_1) { e_1 = { error: e_1_1 }; }
    finally {
        try {
            if (_c && !_c.done && (_a = _b["return"])) _a.call(_b);
        }
        finally { if (e_1) throw e_1.error; }
    }
    return newMatrix;
}
/** Create an iterator over the cells in the matrix */
function entries$2(matrix) {
    var _a, _b, _c, row, values, _d, _e, _f, column, value, point, e_2_1, e_3_1;
    var e_3, _g, e_2, _h;
    return __generator(this, function (_j) {
        switch (_j.label) {
            case 0:
                _j.trys.push([0, 11, 12, 13]);
                _a = __values(matrix.entries()), _b = _a.next();
                _j.label = 1;
            case 1:
                if (!!_b.done) return [3 /*break*/, 10];
                _c = __read(_b.value, 2), row = _c[0], values = _c[1];
                _j.label = 2;
            case 2:
                _j.trys.push([2, 7, 8, 9]);
                _d = (e_2 = void 0, __values(values.entries())), _e = _d.next();
                _j.label = 3;
            case 3:
                if (!!_e.done) return [3 /*break*/, 6];
                _f = __read(_e.value, 2), column = _f[0], value = _f[1];
                point = { row: row, column: column };
                return [4 /*yield*/, [point, value]];
            case 4:
                _j.sent();
                _j.label = 5;
            case 5:
                _e = _d.next();
                return [3 /*break*/, 3];
            case 6: return [3 /*break*/, 9];
            case 7:
                e_2_1 = _j.sent();
                e_2 = { error: e_2_1 };
                return [3 /*break*/, 9];
            case 8:
                try {
                    if (_e && !_e.done && (_h = _d["return"])) _h.call(_d);
                }
                finally { if (e_2) throw e_2.error; }
                return [7 /*endfinally*/];
            case 9:
                _b = _a.next();
                return [3 /*break*/, 1];
            case 10: return [3 /*break*/, 13];
            case 11:
                e_3_1 = _j.sent();
                e_3 = { error: e_3_1 };
                return [3 /*break*/, 13];
            case 12:
                try {
                    if (_b && !_b.done && (_g = _a["return"])) _g.call(_a);
                }
                finally { if (e_3) throw e_3.error; }
                return [7 /*endfinally*/];
            case 13: return [2 /*return*/];
        }
    });
}
/**
 * Converts all elements in row into a string separated by horizontalSeparator and each row string
 * to string separated by verticalSeparator
 */
function join(matrix, horizontalSeparator, verticalSeparator) {
    if (horizontalSeparator === void 0) { horizontalSeparator = "\t"; }
    if (verticalSeparator === void 0) { verticalSeparator = "\n"; }
    var joined = "";
    var _a = getSize(matrix), rows = _a.rows, columns = _a.columns;
    for (var row = 0; row < rows; row++) {
        if (row) {
            joined += verticalSeparator;
        }
        for (var column = 0; column < columns; column++) {
            if (column) {
                joined += horizontalSeparator;
            }
            if (matrix[row] && column in matrix[row]) {
                joined += String(matrix[row][column]);
            }
        }
    }
    return joined;
}
/**
 * Parses a CSV separated by a horizontalSeparator and verticalSeparator into a
 * Matrix using a transform function
 */
function split(csv, transform, horizontalSeparator, verticalSeparator) {
    if (horizontalSeparator === void 0) { horizontalSeparator = "\t"; }
    if (verticalSeparator === void 0) { verticalSeparator = /\r\n|\n|\r/; }
    return csv
        .split(verticalSeparator)
        .map(function (row) { return row.split(horizontalSeparator).map(transform); });
}
/** Returns whether the point exists in the matrix or not. */
function has$3(point, matrix) {
    var firstRow = matrix[0];
    return (firstRow &&
        // validation
        point.row >= 0 &&
        point.column >= 0 &&
        Number.isInteger(point.row) &&
        Number.isInteger(point.column) &&
        // first row length is in sync with other rows
        point.column < firstRow.length &&
        point.row < matrix.length);
}
/** Gets the count of rows and columns of given matrix */
function getSize(matrix) {
    return {
        columns: getColumnsCount(matrix),
        rows: getRowsCount(matrix)
    };
}
/** Gets the count of rows of given matrix */
function getRowsCount(matrix) {
    return matrix.length;
}
/** Gets the count of columns of given matrix */
function getColumnsCount(matrix) {
    var firstRow = matrix[0];
    return firstRow ? firstRow.length : 0;
}
/**
 * Pads matrix with empty columns to match given total columns
 * @param matrix - matrix to pad
 * @param size - minimum size of the matrix after padding.
 * @returns the updated matrix
 */
function pad(matrix, size) {
    var _a = getSize(matrix), rows = _a.rows, columns = _a.columns;
    if (rows >= size.rows && columns >= size.columns) {
        // Optimization, no padding required.
        return matrix;
    }
    var resultSize = {
        rows: size.rows > rows ? size.rows : rows,
        columns: size.columns > columns ? size.columns : columns
    };
    var padded = __spreadArray([], __read(matrix));
    if (resultSize.columns > columns) {
        var padColumns_1 = resultSize.columns - columns;
        padded = padded.map(function (row) { return __spreadArray(__spreadArray([], __read(row), false), __read(Array(padColumns_1).fill(undefined))); });
    }
    if (resultSize.rows > rows) {
        var padRows_1 = resultSize.rows - rows;
        var emptyRow = Array(resultSize.columns).fill(undefined);
        padded = __spreadArray(__spreadArray([], __read(padded), false), __read(Array(padRows_1).fill(emptyRow)));
    }
    return padded;
}
/**
 * Flattens a matrix values to an array
 * @param matrix - the matrix to flatten values from
 * @param transform - optional transform function to apply to each value in the
 * matrix
 * @returns an array of the values from matrix, transformed if a transform
 * function is passed
 */
// eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
function toArray$1(matrix, transform) {
    var array = [];
    for (var row = 0; row < matrix.length; row++) {
        for (var column = 0; column < matrix[row].length; column++) {
            var value = matrix[row][column];
            array.push(transform ? transform(value, { row: row, column: column }) : value);
        }
    }
    return array;
}
/** Returns the maximum point in the matrix */
function maxPoint(matrix) {
    var size = getSize(matrix);
    return { row: size.rows - 1, column: size.columns - 1 };
}

/** Returns whether given value is a point */
function is$1(value) {
    return (
    // @ts-ignore
    typeof (value === null || value === void 0 ? void 0 : value.row) === "number" &&
        // @ts-ignore
        typeof (value === null || value === void 0 ? void 0 : value.column) === "number");
}
/** Return whether two given points are the equal */
function isEqual(source, target) {
    return source.column === target.column && source.row === target.row;
}
/** The origin point in matrices */
var ORIGIN = { row: 0, column: 0 };

/**
 * Interface for ranges between two points
 */
/** Creates a normalized range between two given points */
function create(source, target) {
    return {
        start: {
            row: Math.min(source.row, target.row),
            column: Math.min(source.column, target.column)
        },
        end: {
            row: Math.max(source.row, target.row),
            column: Math.max(source.column, target.column)
        }
    };
}
/** Iterates through all the existing points in given range */
function iterate(range) {
    var row, column;
    return __generator(this, function (_a) {
        switch (_a.label) {
            case 0:
                row = range.start.row;
                _a.label = 1;
            case 1:
                if (!(row <= range.end.row)) return [3 /*break*/, 6];
                column = range.start.column;
                _a.label = 2;
            case 2:
                if (!(column <= range.end.column)) return [3 /*break*/, 5];
                return [4 /*yield*/, { row: row, column: column }];
            case 3:
                _a.sent();
                _a.label = 4;
            case 4:
                column++;
                return [3 /*break*/, 2];
            case 5:
                row++;
                return [3 /*break*/, 1];
            case 6: return [2 /*return*/];
        }
    });
}
/** Returns the size (rows x columns) of the given range */
function size$3(range) {
    var rows = range.end.row + 1 - range.start.row;
    var columns = range.end.column + 1 - range.start.column;
    return rows * columns;
}
/** Returns whether given point exists in given range */
function has$2(range, point) {
    return (point.row >= range.start.row &&
        point.column >= range.start.column &&
        point.row <= range.end.row &&
        point.column <= range.end.column);
}
/** Limits given masked range with given mask */
function mask(masked, mask) {
    return create({
        row: mask.start.row > masked.start.row ? mask.start.row : masked.start.row,
        column: mask.start.column > masked.start.column
            ? mask.start.column
            : masked.start.column
    }, {
        row: mask.end.row < masked.end.row ? mask.end.row : masked.end.row,
        column: mask.end.column < masked.end.column
            ? mask.end.column
            : masked.end.column
    });
}
/** Returns whether given value is a point range */
function is(value) {
    // @ts-ignore
    return is$1(value === null || value === void 0 ? void 0 : value.start) && is$1(value === null || value === void 0 ? void 0 : value.end);
}

var Direction;
(function (Direction) {
    Direction["Left"] = "Left";
    Direction["Right"] = "Right";
    Direction["Top"] = "Top";
    Direction["Bottom"] = "Bottom";
})(Direction || (Direction = {}));
/** Type entirely selected */
var EntireType;
(function (EntireType) {
    EntireType["Row"] = "row";
    EntireType["Column"] = "column";
    EntireType["Table"] = "table";
})(EntireType || (EntireType = {}));
/** Returns whether given selection is entire rows */
function isEntireRows(selection) {
    return (selection !== null &&
        "type" in selection &&
        selection.type === EntireType.Row);
}
/** Returns whether given selection is entire columns */
function isEntireColumns(selection) {
    return (selection !== null &&
        "type" in selection &&
        selection.type === EntireType.Column);
}
/** Returns whether given selection is entire table */
function isEntireTable(selection) {
    return (selection !== null &&
        "type" in selection &&
        selection.type === EntireType.Table);
}
/**
 * Creates entire rows selection
 * @param start - row index where the selection starts, integer
 * @param end - row index where the selection ends, integer
 * @throws {@link InvalidIndexError}
 */
function createEntireRows(start, end) {
    if (!isIndex(start)) {
        throw new InvalidIndexError("start");
    }
    if (!isIndex(end)) {
        throw new InvalidIndexError("end");
    }
    return {
        type: EntireType.Row,
        start: Math.min(start, end),
        end: Math.max(start, end)
    };
}
/**
 * Creates entire columns selection
 * @param start - column index where the selection starts, integer
 * @param end - column index where the selection starts, integer
 */
function createEntireColumns(start, end) {
    if (!isIndex(start)) {
        throw new InvalidIndexError("start");
    }
    if (!isIndex(end)) {
        throw new InvalidIndexError("end");
    }
    return {
        type: EntireType.Column,
        start: Math.min(start, end),
        end: Math.max(start, end)
    };
}
/** Creates entire table selection */
function createEntireTable() {
    // create an object as in the future this might hold the worksheet number
    return { type: EntireType.Table };
}
/** Get concrete range in given data of given selection */
function toRange$1(selection, data) {
    if (selection === null) {
        return null;
    }
    if (is(selection)) {
        return selection;
    }
    switch (selection.type) {
        case EntireType.Row: {
            var max = maxPoint(data);
            return create({ row: selection.start, column: 0 }, { row: selection.end, column: max.column });
        }
        case EntireType.Column: {
            var max = maxPoint(data);
            return create({ row: 0, column: selection.start }, { row: max.row, column: selection.end });
        }
        case EntireType.Table: {
            return getMatrixRange(data);
        }
    }
}
/** Get the number of points selected */
function size$2(selection, data) {
    var range = toRange$1(selection, data);
    return range ? size$3(range) : 0;
}
/** Return whether the given point is within given selection */
function hasPoint(selection, data, point) {
    var range = toRange$1(selection, data);
    return range !== null && has$2(range, point);
}
/** Return whether the given row is entirely selected in given selection */
function hasEntireRow(selection, row) {
    return (isEntireRows(selection) && row >= selection.start && row <= selection.end);
}
/** Return whether the given column is entirely selected in given selection */
function hasEntireColumn(selection, column) {
    return (isEntireColumns(selection) &&
        column >= selection.start &&
        column <= selection.end);
}
/** Normalize given selection to given data matrix */
function normalize(selection, data) {
    if (selection) {
        if (is(selection)) {
            return normalizeRange(selection, data);
        }
        switch (selection.type) {
            case EntireType.Row: {
                return normalizeEntireRows(selection, data);
            }
            case EntireType.Column: {
                return normalizeEntireColumns(selection, data);
            }
        }
    }
    return selection;
}
/** Normalize given range to given data matrix */
function normalizeRange(range, data) {
    var dataRange = getMatrixRange(data);
    return mask(range, dataRange);
}
/** Normalize given entire rows selection to given data matrix */
function normalizeEntireRows(selection, data) {
    var count = getRowsCount(data);
    return createEntireRows(Math.max(selection.start, 0), Math.min(selection.end, count - 1));
}
/** Normalize given entire columns selection to given data matrix */
function normalizeEntireColumns(selection, data) {
    var count = getColumnsCount(data);
    return createEntireColumns(Math.max(selection.start, 0), Math.min(selection.end, count - 1));
}
/** Get selected points */
function getPoints(selection, data) {
    var range = toRange$1(selection, data);
    return range ? Array.from(iterate(range)) : [];
}
/** Get given selection from given data */
function getSelectionFromMatrix(selection, data) {
    var range = toRange$1(selection, data);
    return range ? getRangeFromMatrix(range, data) : [];
}
/** Modify given edge of given selection according to given active point and data matrix */
function modifyEdge(selection, active, data, edge) {
    if (active && selection) {
        if (is(selection)) {
            return modifyRangeEdge(selection, active, data, edge);
        }
        switch (selection.type) {
            case EntireType.Row: {
                return modifyEntireRowsEdge(selection, active, data, edge);
            }
            case EntireType.Column: {
                return modifyEntireColumnsEdge(selection, active, data, edge);
            }
        }
    }
    return selection;
}
/** Modify given edge of given entire rows selection according to given active point and active matrix */
function modifyEntireRowsEdge(selection, active, data, edge) {
    var _a, _b;
    if (edge === Direction.Left || edge === Direction.Right) {
        return selection;
    }
    var delta = edge === Direction.Top ? -1 : 1;
    var property = edge === Direction.Top ? "start" : "end";
    var oppositeProperty = property === "start" ? "end" : "start";
    var nextSelection;
    if (edge === Direction.Top
        ? selection.end > active.row
        : selection.start < active.row) {
        nextSelection = __assign(__assign({}, selection), (_a = {}, _a[oppositeProperty] = selection[oppositeProperty] + delta, _a));
    }
    else {
        nextSelection = __assign(__assign({}, selection), (_b = {}, _b[property] = selection[property] + delta, _b));
    }
    return normalizeEntireRows(nextSelection, data);
}
/** Modify given edge of given entire rows selection according to given active point and active matrix */
function modifyEntireColumnsEdge(selection, active, data, edge) {
    var _a, _b;
    if (edge === Direction.Top || edge === Direction.Bottom) {
        return selection;
    }
    var delta = edge === Direction.Left ? -1 : 1;
    var property = edge === Direction.Left ? "start" : "end";
    var oppositeProperty = property === "start" ? "end" : "start";
    var nextSelection;
    if (edge === Direction.Left
        ? selection.end > active.column
        : selection.start < active.column) {
        nextSelection = __assign(__assign({}, selection), (_a = {}, _a[oppositeProperty] = selection[oppositeProperty] + delta, _a));
    }
    else {
        nextSelection = __assign(__assign({}, selection), (_b = {}, _b[property] = selection[property] + delta, _b));
    }
    return normalizeEntireColumns(nextSelection, data);
}
/** Modify given edge of given range according to given active point and data matrix */
function modifyRangeEdge(range, active, data, edge) {
    var _a, _b, _c;
    var field = edge === Direction.Left || edge === Direction.Right ? "column" : "row";
    var key = edge === Direction.Left || edge === Direction.Top ? "start" : "end";
    var delta = key === "start" ? -1 : 1;
    var edgeOffsets = has$2(range, __assign(__assign({}, active), (_a = {}, _a[field] = active[field] + delta * -1, _a)));
    var keyToModify = edgeOffsets ? (key === "start" ? "end" : "start") : key;
    var nextRange = __assign(__assign({}, range), (_b = {}, _b[keyToModify] = __assign(__assign({}, range[keyToModify]), (_c = {}, _c[field] = range[keyToModify][field] + delta, _c)), _b));
    return normalizeRange(nextRange, data);
}
/** Get the point range of given matrix */
function getMatrixRange(data) {
    var maxPoint$1 = maxPoint(data);
    return create(ORIGIN, maxPoint$1);
}
/** Get a matrix of values in given range from given matrix */
function getRangeFromMatrix(range, matrix) {
    return slice(range.start, range.end, matrix);
}
/** Returns whether given value is a valid index */
function isIndex(value) {
    return Number.isInteger(value) && value >= 0;
}
/** Error thrown when passing a non-index value where index is expected */
var InvalidIndexError = /** @class */ (function (_super) {
    __extends(InvalidIndexError, _super);
    function InvalidIndexError(name) {
        return _super.call(this, "".concat(name, " is not a valid index. It must be 0 or a positive integer")) || this;
    }
    return InvalidIndexError;
}(Error));

/** Sets the value for point in map */
function set$1(point, value, map) {
    var _a, _b;
    return __assign(__assign({}, map), (_a = {}, _a[point.row] = __assign(__assign({}, map[point.row]), (_b = {}, _b[point.column] = value, _b)), _a));
}
function unset(_a, map) {
    var _b;
    var row = _a.row, column = _a.column;
    if (!(row in map) || !(column in map[row])) {
        return map;
    }
    var _c = map, 
    // @ts-ignore
    _d = String(row), 
    // @ts-ignore
    _e = _c[_d], _f = String(column); _e[_f]; var nextRow = __rest(_e, [typeof _f === "symbol" ? _f : _f + ""]), nextMap = __rest(_c, [typeof _d === "symbol" ? _d : _d + ""]);
    if (Object.keys(nextRow).length === 0) {
        return nextMap;
    }
    return __assign(__assign({}, nextMap), (_b = {}, _b[row] = nextRow, _b));
}
/** Gets the value for point in map */
function get(point, map) {
    return map[point.row] && map[point.row][point.column];
}
/** Checks if map has point assigned to value */
function has$1(point, map) {
    return point.row in map && point.column in map[point.row];
}
var EMPTY = {};
/** Creates a new PointMap instance from an array-like or iterable object. */
function from$2(pairs) {
    return pairs.reduce(function (acc, _a) {
        var _b = __read(_a, 2), point = _b[0], value = _b[1];
        return set$1(point, value, acc);
    }, EMPTY);
}
/** Creates a new PointMap instance from a Matrix. */
function fromMatrix(matrix) {
    return matrix.reduce(function (rowAcc, data, row) {
        return data.reduce(function (colAcc, cell, column) {
            return cell ? set$1({ row: row, column: column }, cell, colAcc) : colAcc;
        }, rowAcc);
    }, EMPTY);
}
/** Returns the number of elements in a PointMap object. */
function size$1(map) {
    var acc = 0;
    var _map_keys = Object.keys(map);
    for (var i = 0; i < _map_keys.length; i++) {
        var row = Number(_map_keys[i]);
        var columns = map[row];
        acc += Object.keys(columns).length;
    }
    return acc;
}
/** Applies a function against an accumulator and each value and point in the map (from left to right) to reduce it to a single value */
function reduce(func, map, initialValue) {
    var acc = initialValue;
    var _map_keys = Object.keys(map);
    for (var i = 0; i < _map_keys.length; i++) {
        var row = Number(_map_keys[i]);
        var columns = map[row];
        var _columns_keys = Object.keys(columns);
        for (var j = 0; j < _columns_keys.length; j++) {
            var column = Number(_columns_keys[j]);
            var value = columns[column];
            acc = func(acc, value, { row: row, column: column });
        }
    }
    return acc;
}
/** Creates a new map with the results of calling a provided function on every value in the calling map */
function map(func, map) {
    return reduce(function (acc, value, point) { return set$1(point, func(value), acc); }, map, from$2([]));
}
/** Returns whether map has any points set to value */
function isEmpty(map) {
    return Object.keys(map).length === 0;
}
function entries$1(map) {
    var _a, _b, _c, _i, row, _d, _e, _f, _g, column;
    return __generator(this, function (_h) {
        switch (_h.label) {
            case 0:
                _a = map;
                _b = [];
                for (_c in _a)
                    _b.push(_c);
                _i = 0;
                _h.label = 1;
            case 1:
                if (!(_i < _b.length)) return [3 /*break*/, 6];
                _c = _b[_i];
                if (!(_c in _a)) return [3 /*break*/, 5];
                row = _c;
                _d = map[row];
                _e = [];
                for (_f in _d)
                    _e.push(_f);
                _g = 0;
                _h.label = 2;
            case 2:
                if (!(_g < _e.length)) return [3 /*break*/, 5];
                _f = _e[_g];
                if (!(_f in _d)) return [3 /*break*/, 4];
                column = _f;
                return [4 /*yield*/, [{ row: Number(row), column: Number(column) }, map[row][column]]];
            case 3:
                _h.sent();
                _h.label = 4;
            case 4:
                _g++;
                return [3 /*break*/, 2];
            case 5:
                _i++;
                return [3 /*break*/, 1];
            case 6: return [2 /*return*/];
        }
    });
}

/**
 * Immutable Set like interface of points
 */
/** Returns a boolean asserting whether an point is present with the given value in the Set object or not */
var has = function (set, point) {
    return has$1(point, set);
};
/** Returns the number of points in a PointSet object */
var size = function (set) { return size$1(set); };
var minKey = function (object) {
    /* @ts-ignore*/
    return Math.min.apply(Math, __spreadArray([], __read(Object.keys(object))));
};
/** Returns the point on the minimal row in the minimal column in the set */
function min(set) {
    var row = minKey(set);
    return { row: row, column: minKey(set[row]) };
}
var maxKey = function (object) {
    // @ts-ignore
    return Math.max.apply(Math, __spreadArray([], __read(Object.keys(object))));
};
/** Returns the point on the maximal row in the maximal column in the set */
function max(set) {
    var row = maxKey(set);
    return { row: row, column: maxKey(set[row]) };
}
/** Creates a new PointSet instance from an array-like or iterable object */
function from$1(points) {
    return points.reduce(function (acc, point) { return set$1(point, true, acc); }, from$2([]));
}
/** Transform a point set to a range */
function toRange(set) {
    var start = min(set);
    var end = max(set);
    return create(start, end);
}
/** Add the given point to given set */
function add(point, set) {
    return set$1(point, true, set);
}
/** Remove the given point to given set */
function remove(point, set) {
    return unset(point, set);
}
/** Create an array from given set */
function toArray(set) {
    return reduce(function (acc, value, point) { return __spreadArray(__spreadArray([], __read(acc), false), [point]); }, set, []);
}
function subtract(toSubtract, set) {
    var e_1, _a;
    var newSet = set;
    try {
        for (var _b = __values(toArray(toSubtract)), _c = _b.next(); !_c.done; _c = _b.next()) {
            var point = _c.value;
            newSet = remove(point, newSet);
        }
    }
    catch (e_1_1) { e_1 = { error: e_1_1 }; }
    finally {
        try {
            if (_c && !_c.done && (_a = _b["return"])) _a.call(_b);
        }
        finally { if (e_1) throw e_1.error; }
    }
    return newSet;
}
function entries(set) {
    var _a, _b, _c, point, e_2_1;
    var e_2, _d;
    return __generator(this, function (_e) {
        switch (_e.label) {
            case 0:
                _e.trys.push([0, 5, 6, 7]);
                _a = __values(entries$1(set)), _b = _a.next();
                _e.label = 1;
            case 1:
                if (!!_b.done) return [3 /*break*/, 4];
                _c = __read(_b.value, 1), point = _c[0];
                return [4 /*yield*/, point];
            case 2:
                _e.sent();
                _e.label = 3;
            case 3:
                _b = _a.next();
                return [3 /*break*/, 1];
            case 4: return [3 /*break*/, 7];
            case 5:
                e_2_1 = _e.sent();
                e_2 = { error: e_2_1 };
                return [3 /*break*/, 7];
            case 6:
                try {
                    if (_b && !_b.done && (_d = _a["return"])) _d.call(_a);
                }
                finally { if (e_2) throw e_2.error; }
                return [7 /*endfinally*/];
            case 7: return [2 /*return*/];
        }
    });
}

var PLAIN_TEXT_MIME = "text/plain";
var FOCUS_WITHIN_SELECTOR = ":focus-within";
/** Move the cursor of given input element to the input's end */
function moveCursorToEnd(el) {
    el.selectionStart = el.selectionEnd = el.value.length;
}
/**
 * Creates an array of numbers (positive and/or negative) progressing from start up to, but not including, end. A step of -1 is used if a negative start is specified without an end or step. If end is not specified, it's set to start with start then set to 0.
 * @param end - an integer number specifying at which position to stop (not included).
 * @param start - An integer number specifying at which position to start.
 * @param step - An integer number specifying the incrementation
 */
function range(end, start, step) {
    if (start === void 0) { start = 0; }
    if (step === void 0) { step = 1; }
    var array = [];
    if (Math.sign(end - start) === -1) {
        for (var element = start; element > end; element -= step) {
            array.push(element);
        }
        return array;
    }
    for (var element = start; element < end; element += step) {
        array.push(element);
    }
    return array;
}
/** Return whether given point is active */
function isActive(active, point) {
    return Boolean(active && isEqual(point, active));
}
/** Get the offset values of given element */
function getOffsetRect(element) {
    return {
        width: element.offsetWidth,
        height: element.offsetHeight,
        left: element.offsetLeft,
        top: element.offsetTop
    };
}
/** Write given data to clipboard with given event */
function writeTextToClipboard(event, data) {
    var _a;
    (_a = event.clipboardData) === null || _a === void 0 ? void 0 : _a.setData(PLAIN_TEXT_MIME, data);
}
/** Read text from given clipboard event */
function readTextFromClipboard(event) {
    // @ts-ignore
    if (window.clipboardData && window.clipboardData.getData) {
        // @ts-ignore
        return window.clipboardData.getData("Text");
    }
    if (event.clipboardData && event.clipboardData.getData) {
        return event.clipboardData.getData(PLAIN_TEXT_MIME);
    }
    return "";
}
/** Get the dimensions of cell at point from state */
function getCellDimensions(point, rowDimensions, columnDimensions) {
    var cellRowDimensions = rowDimensions && rowDimensions[point.row];
    var cellColumnDimensions = columnDimensions && columnDimensions[point.column];
    return (cellRowDimensions &&
        cellColumnDimensions && __assign(__assign({}, cellRowDimensions), cellColumnDimensions));
}
/** Get the dimensions of a range of cells */
function getRangeDimensions(rowDimensions, columnDimensions, range) {
    var startDimensions = getCellDimensions(range.start, rowDimensions, columnDimensions);
    var endDimensions = getCellDimensions(range.end, rowDimensions, columnDimensions);
    return (startDimensions &&
        endDimensions && {
        width: endDimensions.left + endDimensions.width - startDimensions.left,
        height: endDimensions.top + endDimensions.height - startDimensions.top,
        top: startDimensions.top,
        left: startDimensions.left
    });
}
/** Get the dimensions of selected */
function getSelectedDimensions(rowDimensions, columnDimensions, data, selected) {
    var range = toRange$1(selected, data);
    return range
        ? getRangeDimensions(rowDimensions, columnDimensions, range)
        : undefined;
}
/** Get given data as CSV */
function getCSV(data) {
    var valueMatrix = map$1(function (cell) { return (cell === null || cell === void 0 ? void 0 : cell.value) || ""; }, data);
    return join(valueMatrix);
}
/**
 * Calculate the rows and columns counts of a spreadsheet
 * @param data - the spreadsheet's data
 * @param rowLabels - the spreadsheet's row labels (if defined)
 * @param columnLabels - the spreadsheet's column labels (if defined)
 * @returns the rows and columns counts of a spreadsheet
 */
function calculateSpreadsheetSize(data, rowLabels, columnLabels) {
    var _a = getSize(data), columns = _a.columns, rows = _a.rows;
    return {
        rows: rowLabels ? Math.max(rows, rowLabels.length) : rows,
        columns: columnLabels ? Math.max(columns, columnLabels.length) : columns
    };
}
/** Transform given point map to a point set */
function convertPointMapToPointSet(map$1) {
    return map(function () { return true; }, map$1);
}
/** Get the range of copied cells. If none are copied return null */
function getCopiedRange(copied, hasPasted) {
    if (hasPasted || isEmpty(copied)) {
        return null;
    }
    var set = convertPointMapToPointSet(copied);
    return toRange(set);
}
/** Should spreadsheet handle clipboard event */
function shouldHandleClipboardEvent(root, mode) {
    return root !== null && mode === "view" && isFocusedWithin(root);
}
function isFocusedWithin(element) {
    return element.matches(FOCUS_WITHIN_SELECTOR);
}

var Table = function (_a) {
    var children = _a.children, columns = _a.columns, hideColumnIndicators = _a.hideColumnIndicators;
    var columnCount = columns + (hideColumnIndicators ? 0 : 1);
    var columnNodes = range(columnCount).map(function (i) { return React.createElement("col", { key: i }); });
    return (React.createElement("table", { className: "Spreadsheet__table" },
        React.createElement("colgroup", null, columnNodes),
        React.createElement("tbody", null, children)));
};

var Row = function (props) { return React.createElement("tr", __assign({}, props)); };

var HeaderRow = function (props) { return React.createElement("tr", __assign({}, props)); };

var FORMULA_VALUE_PREFIX = "=";
/** Returns whether given value is a formula */
function isFormulaValue(value) {
    return (typeof value === "string" &&
        value.startsWith(FORMULA_VALUE_PREFIX) &&
        value.length > 1);
}
/** Extracts formula from value  */
function extractFormula(value) {
    return value.slice(1);
}
function createBoundFormulaParser(getData) {
    return new FormulaParser({
        onCell: function (ref) {
            var point = {
                row: ref.row - 1,
                column: ref.col - 1
            };
            var cell = get$1(point, getData());
            if (!isNaN(cell === null || cell === void 0 ? void 0 : cell.value))
                return Number(cell === null || cell === void 0 ? void 0 : cell.value);
            return cell === null || cell === void 0 ? void 0 : cell.value;
        },
        onRange: function (ref) {
            var data = getData();
            var size = getSize(data);
            var start = {
                row: ref.from.row - 1,
                column: ref.from.col - 1
            };
            var end = {
                row: Math.min(ref.to.row - 1, size.rows - 1),
                column: Math.min(ref.to.col - 1, size.columns - 1)
            };
            var dataSlice = slice(start, end, data);
            return toArray$1(dataSlice, function (cell) {
                if (!isNaN(cell === null || cell === void 0 ? void 0 : cell.value))
                    return Number(cell === null || cell === void 0 ? void 0 : cell.value);
                return cell === null || cell === void 0 ? void 0 : cell.value;
            });
        }
    });
}
var depParser = new DepParser();
/**
 * For given formula returns the cell references
 * @param formula - formula to get references for
 */
function getReferences(formula, point, data) {
    var _a = getSize(data), rows = _a.rows, columns = _a.columns;
    try {
        var dependencies = depParser.parse(formula, convertPointToCellRef(point));
        var references = from$1(dependencies.flatMap(function (reference) {
            var isRange = "from" in reference;
            if (isRange) {
                var from = reference.from, to = reference.to;
                var normalizedFrom = {
                    row: from.row - 1,
                    column: from.col - 1
                };
                var normalizedTo = {
                    row: Math.min(to.row - 1, rows - 1),
                    column: Math.min(to.col - 1, columns - 1)
                };
                var range = create(normalizedFrom, normalizedTo);
                return Array.from(iterate(range));
            }
            return { row: reference.row - 1, column: reference.col - 1 };
        }));
        return references;
    }
    catch (error) {
        console.error(error);
        if (error instanceof FormulaError) {
            return from$1([]);
        }
        else {
            throw error;
        }
    }
}
function evaluate(formula, point, formulaParser) {
    try {
        var position = convertPointToCellRef(point);
        var returned = formulaParser.parse(formula, position);
        return returned instanceof FormulaError ? returned.toString() : returned;
    }
    catch (error) {
        if (error instanceof FormulaError) {
            return error.toString();
        }
        throw error;
    }
}
function convertPointToCellRef(point) {
    return {
        row: point.row + 1,
        col: point.column + 1,
        // TODO: fill once we support multiple sheets
        sheet: "Sheet1"
    };
}

function from(pairs) {
    var e_1, _a, e_2, _b;
    var backward = from$2([]);
    try {
        for (var pairs_1 = __values(pairs), pairs_1_1 = pairs_1.next(); !pairs_1_1.done; pairs_1_1 = pairs_1.next()) {
            var _c = __read(pairs_1_1.value, 2), point = _c[0], points = _c[1];
            var set_1 = get(point, backward) || from$1([]);
            try {
                for (var _d = (e_2 = void 0, __values(toArray(points))), _e = _d.next(); !_e.done; _e = _d.next()) {
                    var p = _e.value;
                    backward = set$1(p, add(point, set_1), backward);
                }
            }
            catch (e_2_1) { e_2 = { error: e_2_1 }; }
            finally {
                try {
                    if (_e && !_e.done && (_b = _d["return"])) _b.call(_d);
                }
                finally { if (e_2) throw e_2.error; }
            }
        }
    }
    catch (e_1_1) { e_1 = { error: e_1_1 }; }
    finally {
        try {
            if (pairs_1_1 && !pairs_1_1.done && (_a = pairs_1["return"])) _a.call(pairs_1);
        }
        finally { if (e_1) throw e_1.error; }
    }
    return {
        forward: from$2(pairs),
        backward: backward
    };
}
function set(point, points, graph) {
    var e_3, _a, e_4, _b;
    var newForward = size(points) === 0
        ? unset(point, graph.forward)
        : set$1(point, points, graph.forward);
    var existing = get(point, graph.forward);
    var toAdd = existing ? subtract(existing, points) : points;
    var newBackward = graph.backward;
    try {
        for (var _c = __values(toArray(toAdd)), _d = _c.next(); !_d.done; _d = _c.next()) {
            var p = _d.value;
            var set_2 = get(p, newBackward) || from$1([]);
            newBackward = set$1(p, add(point, set_2), newBackward);
        }
    }
    catch (e_3_1) { e_3 = { error: e_3_1 }; }
    finally {
        try {
            if (_d && !_d.done && (_a = _c["return"])) _a.call(_c);
        }
        finally { if (e_3) throw e_3.error; }
    }
    if (existing) {
        var toRemove = subtract(points, existing);
        try {
            for (var _e = __values(toArray(toRemove)), _f = _e.next(); !_f.done; _f = _e.next()) {
                var p = _f.value;
                var set_3 = get(p, newBackward);
                if (!set_3) {
                    continue;
                }
                var newSet = remove(point, set_3);
                if (size(newSet) === 0) {
                    newBackward = unset(p, newBackward);
                }
                else {
                    newBackward = set$1(p, newSet, newBackward);
                }
            }
        }
        catch (e_4_1) { e_4 = { error: e_4_1 }; }
        finally {
            try {
                if (_f && !_f.done && (_b = _e["return"])) _b.call(_e);
            }
            finally { if (e_4) throw e_4.error; }
        }
    }
    return {
        forward: newForward,
        backward: newBackward
    };
}
function getBackwardsRecursive(point, graph) {
    var stack, current, backwardDependencies, _a, _b, dependent, e_5_1;
    var e_5, _c;
    return __generator(this, function (_d) {
        switch (_d.label) {
            case 0:
                stack = [point];
                _d.label = 1;
            case 1:
                if (!(stack.length > 0)) return [3 /*break*/, 10];
                current = stack.pop();
                if (!current) {
                    return [3 /*break*/, 1];
                }
                backwardDependencies = get(current, graph.backward);
                // If there are no dependents, skip to the next point
                if (!backwardDependencies) {
                    return [3 /*break*/, 1];
                }
                _d.label = 2;
            case 2:
                _d.trys.push([2, 7, 8, 9]);
                _a = (e_5 = void 0, __values(entries(backwardDependencies))), _b = _a.next();
                _d.label = 3;
            case 3:
                if (!!_b.done) return [3 /*break*/, 6];
                dependent = _b.value;
                return [4 /*yield*/, dependent];
            case 4:
                _d.sent();
                stack.push(dependent);
                _d.label = 5;
            case 5:
                _b = _a.next();
                return [3 /*break*/, 3];
            case 6: return [3 /*break*/, 9];
            case 7:
                e_5_1 = _d.sent();
                e_5 = { error: e_5_1 };
                return [3 /*break*/, 9];
            case 8:
                try {
                    if (_b && !_b.done && (_c = _a["return"])) _c.call(_a);
                }
                finally { if (e_5) throw e_5.error; }
                return [7 /*endfinally*/];
            case 9: return [3 /*break*/, 1];
            case 10: return [2 /*return*/];
        }
    });
}
function hasCircularDependency(graph, startPoint) {
    var e_6, _a;
    var visited = from$1([]);
    var stack = [startPoint];
    while (stack.length > 0) {
        var current = stack.pop();
        if (!current) {
            continue;
        }
        if (has(visited, current)) {
            return true;
        }
        visited = add(current, visited);
        var dependents = get(current, graph.forward);
        if (!dependents) {
            continue;
        }
        try {
            for (var _b = (e_6 = void 0, __values(entries(dependents))), _c = _b.next(); !_c.done; _c = _b.next()) {
                var dependent = _c.value;
                stack.push(dependent);
            }
        }
        catch (e_6_1) { e_6 = { error: e_6_1 }; }
        finally {
            try {
                if (_c && !_c.done && (_a = _b["return"])) _a.call(_b);
            }
            finally { if (e_6) throw e_6.error; }
        }
    }
    return false;
}
function traverseBFS(graph) {
    var visited, queue, _a, _b, _c, point, dependencies, point, dependents, _d, _e, dependent;
    var e_7, _f, e_8, _g;
    return __generator(this, function (_h) {
        switch (_h.label) {
            case 0:
                visited = from$1([]);
                queue = [];
                try {
                    // Iterate over all the points in the forward map and add the ones with no dependencies to the queue
                    for (_a = __values(entries$1(graph.forward)), _b = _a.next(); !_b.done; _b = _a.next()) {
                        _c = __read(_b.value, 2), point = _c[0], dependencies = _c[1];
                        if (size(dependencies) === 0 && !has(visited, point)) {
                            queue.push(point);
                            visited = add(point, visited);
                        }
                    }
                }
                catch (e_7_1) { e_7 = { error: e_7_1 }; }
                finally {
                    try {
                        if (_b && !_b.done && (_f = _a["return"])) _f.call(_a);
                    }
                    finally { if (e_7) throw e_7.error; }
                }
                _h.label = 1;
            case 1:
                if (!(queue.length > 0)) return [3 /*break*/, 3];
                point = queue.shift();
                if (!point) {
                    return [3 /*break*/, 1];
                }
                return [4 /*yield*/, point];
            case 2:
                _h.sent();
                dependents = get(point, graph.forward);
                // If there are no dependents, skip to the next iteration
                if (!dependents) {
                    return [3 /*break*/, 1];
                }
                try {
                    // Otherwise, add the dependents to the queue if they have not yet been visited
                    for (_d = (e_8 = void 0, __values(entries(dependents))), _e = _d.next(); !_e.done; _e = _d.next()) {
                        dependent = _e.value;
                        if (!has(visited, dependent)) {
                            queue.push(dependent);
                            visited = add(dependent, visited);
                        }
                    }
                }
                catch (e_8_1) { e_8 = { error: e_8_1 }; }
                finally {
                    try {
                        if (_e && !_e.done && (_g = _d["return"])) _g.call(_d);
                    }
                    finally { if (e_8) throw e_8.error; }
                }
                return [3 /*break*/, 1];
            case 3: return [2 /*return*/];
        }
    });
}

var Model = /** @class */ (function () {
    function Model(data, referenceGraph, evaluatedData) {
        this.data = data;
        this.referenceGraph = referenceGraph || createReferenceGraph(data);
        this.evaluatedData =
            evaluatedData || createEvaluatedData(data, this.referenceGraph);
    }
    return Model;
}());
function updateCellValue(model, point, cell) {
    var nextData = set$2(point, cell, model.data);
    var nextReferenceGraph = isFormulaValue(cell.value)
        ? updateReferenceGraph(model.referenceGraph, point, cell, nextData)
        : model.referenceGraph;
    var nextEvaluatedData = evaluateCell(model.evaluatedData, nextData, nextReferenceGraph, point, cell);
    return new Model(nextData, nextReferenceGraph, nextEvaluatedData);
}
function updateReferenceGraph(referenceGraph, point, cell, data) {
    var references = getReferences(extractFormula(cell.value), point, data);
    var nextReferenceGraph = set(point, references, referenceGraph);
    return nextReferenceGraph;
}
function evaluateCell(prevEvaluatedData, data, referenceGraph, point, cell) {
    var e_1, _a, e_2, _b;
    if (hasCircularDependency(referenceGraph, point)) {
        var visited = from$1([point]);
        var nextEvaluatedData_1 = set$2(point, __assign(__assign({}, cell), { value: "#REF!" }), prevEvaluatedData);
        try {
            for (var _c = __values(getBackwardsRecursive(point, referenceGraph)), _d = _c.next(); !_d.done; _d = _c.next()) {
                var referrer = _d.value;
                if (has(visited, referrer)) {
                    break;
                }
                visited = add(referrer, visited);
                var referrerCell = get$1(referrer, data);
                if (!referrerCell) {
                    continue;
                }
                nextEvaluatedData_1 = set$2(referrer, __assign(__assign({}, referrerCell), { value: "#REF!" }), nextEvaluatedData_1);
            }
        }
        catch (e_1_1) { e_1 = { error: e_1_1 }; }
        finally {
            try {
                if (_d && !_d.done && (_a = _c["return"])) _a.call(_c);
            }
            finally { if (e_1) throw e_1.error; }
        }
        return nextEvaluatedData_1;
    }
    var nextEvaluatedData = prevEvaluatedData;
    var formulaParser = createBoundFormulaParser(function () { return nextEvaluatedData; });
    var evaluatedValue = isFormulaValue(cell.value)
        ? getFormulaComputedValue(cell.value, point, formulaParser)
        : cell.value;
    var evaluatedCell = __assign(__assign({}, cell), { value: evaluatedValue });
    nextEvaluatedData = set$2(point, evaluatedCell, nextEvaluatedData);
    try {
        // for every formula cell that references the cell re-evaluate (recursive)
        for (var _e = __values(getBackwardsRecursive(point, referenceGraph)), _f = _e.next(); !_f.done; _f = _e.next()) {
            var referrer = _f.value;
            var referrerCell = get$1(referrer, data);
            if (!referrerCell) {
                continue;
            }
            var evaluatedValue_1 = isFormulaValue(referrerCell.value)
                ? getFormulaComputedValue(referrerCell.value, point, formulaParser)
                : referrerCell.value;
            var evaluatedCell_1 = __assign(__assign({}, referrerCell), { value: evaluatedValue_1 });
            nextEvaluatedData = set$2(referrer, evaluatedCell_1, nextEvaluatedData);
        }
    }
    catch (e_2_1) { e_2 = { error: e_2_1 }; }
    finally {
        try {
            if (_f && !_f.done && (_b = _e["return"])) _b.call(_e);
        }
        finally { if (e_2) throw e_2.error; }
    }
    return nextEvaluatedData;
}
/**
 *
 * @param data - the spreadsheet data
 * @returns the spreadsheet reference graph
 */
function createReferenceGraph(data) {
    var e_3, _a;
    var entries = [];
    try {
        for (var _b = __values(entries$2(data)), _c = _b.next(); !_c.done; _c = _b.next()) {
            var _d = __read(_c.value, 2), point = _d[0], cell = _d[1];
            if (cell && isFormulaValue(cell.value)) {
                var references = getReferences(extractFormula(cell.value), point, data);
                entries.push([point, references]);
            }
        }
    }
    catch (e_3_1) { e_3 = { error: e_3_1 }; }
    finally {
        try {
            if (_c && !_c.done && (_a = _b["return"])) _a.call(_b);
        }
        finally { if (e_3) throw e_3.error; }
    }
    return from(entries);
}
function createEvaluatedData(data, referenceGraph) {
    var e_4, _a;
    var evaluatedData = data;
    var formulaParser = createBoundFormulaParser(function () { return evaluatedData; });
    try {
        // Iterate over the points in the reference graph, starting from the leaves
        for (var _b = __values(traverseBFS(referenceGraph)), _c = _b.next(); !_c.done; _c = _b.next()) {
            var point = _c.value;
            // Get the cell at the current point in the data matrix
            var cell = get$1(point, data);
            if (!cell) {
                continue;
            }
            // If the cell is a formula cell, evaluate it
            if (isFormulaValue(cell.value)) {
                var evaluatedValue = getFormulaComputedValue(cell.value, point, formulaParser);
                evaluatedData = set$2(point, __assign(__assign({}, cell), { value: evaluatedValue }), evaluatedData);
            }
        }
    }
    catch (e_4_1) { e_4 = { error: e_4_1 }; }
    finally {
        try {
            if (_c && !_c.done && (_a = _b["return"])) _a.call(_b);
        }
        finally { if (e_4) throw e_4.error; }
    }
    return evaluatedData;
}
/** Get the computed value of a formula cell */
function getFormulaComputedValue(value, point, formulaParser) {
    var formula = extractFormula(value);
    return evaluate(formula, point, formulaParser);
}

var INITIAL_STATE = {
    active: null,
    mode: "view",
    rowDimensions: {},
    columnDimensions: {},
    lastChanged: null,
    hasPasted: false,
    cut: false,
    dragging: false,
    model: new Model([]),
    selected: null,
    copied: from$2([]),
    lastCommit: null
};
var reducer = createReducer(INITIAL_STATE, function (builder) {
    builder.addCase(setData, function (state, action) {
        var data = action.payload.data;
        var nextActive = state.active && has$3(state.active, data) ? state.active : null;
        var nextSelected = normalize(state.selected, data);
        return __assign(__assign({}, state), { model: new Model(data), active: nextActive, selected: nextSelected });
    });
    builder.addCase(select, function (state, action) {
        var point = action.payload.point;
        if (state.active && !isActive(state.active, point)) {
            return __assign(__assign({}, state), { selected: create(point, state.active), mode: "view" });
        }
    });
    builder.addCase(selectEntireTable, function (state) {
        return __assign(__assign({}, state), { selected: createEntireTable(), active: ORIGIN, mode: "view" });
    });
    builder.addCase(selectEntireColumn, function (state, action) {
        var _a = action.payload, column = _a.column, extend = _a.extend;
        var active = state.active;
        return __assign(__assign({}, state), { selected: extend && active
                ? createEntireColumns(active.column, column)
                : createEntireColumns(column, column), active: extend && active ? active : __assign(__assign({}, ORIGIN), { column: column }), mode: "view" });
    });
    builder.addCase(selectEntireRow, function (state, action) {
        var _a = action.payload, row = _a.row, extend = _a.extend;
        var active = state.active;
        return __assign(__assign({}, state), { selected: extend && active
                ? createEntireRows(active.row, row)
                : createEntireRows(row, row), active: extend && active ? active : __assign(__assign({}, ORIGIN), { row: row }), mode: "view" });
    });
    builder.addCase(activate, function (state, action) {
        var point = action.payload.point;
        return __assign(__assign({}, state), { selected: create(point, point), active: point, mode: isActive(state.active, point) ? "edit" : "view" });
    });
    builder.addCase(setCellData, function (state, action) {
        var _a = action.payload, active = _a.active, cellData = _a.data;
        if (isActiveReadOnly(state)) {
            return;
        }
        return __assign(__assign({}, state), { model: updateCellValue(state.model, active, cellData), lastChanged: active });
    });
    builder.addCase(setCellDimensions, function (state, action) {
        var _a, _b;
        var _c = action.payload, point = _c.point, dimensions = _c.dimensions;
        var prevRowDimensions = state.rowDimensions[point.row];
        var prevColumnDimensions = state.columnDimensions[point.column];
        if (prevRowDimensions &&
            prevColumnDimensions &&
            prevRowDimensions.top === dimensions.top &&
            prevRowDimensions.height === dimensions.height &&
            prevColumnDimensions.left === dimensions.left &&
            prevColumnDimensions.width === dimensions.width) {
            return;
        }
        return __assign(__assign({}, state), { rowDimensions: __assign(__assign({}, state.rowDimensions), (_a = {}, _a[point.row] = { top: dimensions.top, height: dimensions.height }, _a)), columnDimensions: __assign(__assign({}, state.columnDimensions), (_b = {}, _b[point.column] = { left: dimensions.left, width: dimensions.width }, _b)) });
    });
    builder.addCase(paste, function (state, action) {
        var text = action.payload.data;
        var active = state.active;
        if (!active) {
            return;
        }
        var copiedMatrix = split(text, function (value) { return ({ value: value }); });
        var copied = fromMatrix(copiedMatrix);
        var minPoint = min(copied);
        var copiedSize = getSize(copiedMatrix);
        var requiredSize = {
            rows: active.row + copiedSize.rows,
            columns: active.column + copiedSize.columns
        };
        var paddedData = pad(state.model.data, requiredSize);
        var _a = reduce(function (acc, value, point) {
            var commit = acc.commit || [];
            var nextPoint = {
                row: point.row - minPoint.row + active.row,
                column: point.column - minPoint.column + active.column
            };
            var nextData = state.cut ? unset$1(point, acc.data) : acc.data;
            if (state.cut) {
                commit = __spreadArray(__spreadArray([], __read(commit), false), [{ prevCell: value, nextCell: null }]);
            }
            if (!has$3(nextPoint, paddedData)) {
                return { data: nextData, commit: commit };
            }
            var currentValue = get$1(nextPoint, nextData) || null;
            commit = __spreadArray(__spreadArray([], __read(commit), false), [
                {
                    prevCell: currentValue,
                    nextCell: value
                },
            ]);
            return {
                data: set$2(nextPoint, __assign(__assign({}, currentValue), value), nextData),
                commit: commit
            };
        }, copied, { data: paddedData, commit: [] }), data = _a.data, commit = _a.commit;
        return __assign(__assign({}, state), { data: data, selected: create(active, {
                row: active.row + copiedSize.rows - 1,
                column: active.column + copiedSize.columns - 1
            }), cut: false, hasPasted: true, mode: "view", lastCommit: commit });
    });
    builder.addCase(edit$1, edit);
    builder.addCase(view$1, view);
    builder.addCase(clear$1, clear);
    builder.addCase(blur$1, blur);
    builder.addCase(keyPress, function (state, action) {
        var event = action.payload.event;
        if (isActiveReadOnly(state) || event.metaKey) {
            return;
        }
        if (state.mode === "view" && state.active) {
            return edit(state);
        }
        return;
    });
    builder.addCase(keyDown, function (state, action) {
        var event = action.payload.event;
        var handler = getKeyDownHandler(state, event);
        if (handler) {
            return __assign(__assign({}, state), handler(state, event));
        }
        return;
    });
    builder.addCase(dragStart, function (state, action) {
        return __assign(__assign({}, state), { dragging: true });
    });
    builder.addCase(dragEnd, function (state, action) {
        return __assign(__assign({}, state), { dragging: false });
    });
    builder.addCase(commit$1, function (state, action) {
        var changes = action.payload.changes;
        return __assign(__assign({}, state), commit(changes));
    });
    builder.addMatcher(function (action) {
        return action.type === copy.type || action.type === cut.type;
    }, function (state, action) {
        var selectedPoints = getPoints(state.selected, state.model.data);
        return __assign(__assign({}, state), { copied: selectedPoints.reduce(function (acc, point) {
                var cell = get$1(point, state.model.data);
                return cell === undefined ? acc : set$1(point, cell, acc);
            }, from$2([])), cut: action.type === cut.type, hasPasted: false });
    });
});
// Shared reducers
function edit(state) {
    if (isActiveReadOnly(state)) {
        return;
    }
    return __assign(__assign({}, state), { mode: "edit" });
}
function clear(state) {
    if (!state.active) {
        return;
    }
    var canClearCell = function (cell) {
        return cell && !cell.readOnly;
    };
    var clearCell = function (cell) {
        if (!canClearCell(cell)) {
            return cell;
        }
        return Object.assign({}, cell, { value: undefined });
    };
    var selectedPoints = getPoints(state.selected, state.model.data);
    var changes = selectedPoints.map(function (point) {
        var cell = get$1(point, state.model.data);
        return __assign(__assign({}, state), { prevCell: cell || null, nextCell: clearCell(cell) || null });
    });
    var newData = selectedPoints.reduce(function (acc, point) {
        var cell = get$1(point, acc);
        return set$2(point, clearCell(cell), acc);
    }, state.model.data);
    return __assign(__assign(__assign({}, state), { model: new Model(newData) }), commit(changes));
}
function blur(state) {
    return __assign(__assign({}, state), { active: null, selected: null });
}
function view(state) {
    return __assign(__assign({}, state), { mode: "view" });
}
function commit(changes) {
    return { lastCommit: changes };
}
// Utility
var go = function (rowDelta, columnDelta) {
    return function (state) {
        if (!state.active) {
            return;
        }
        var nextActive = {
            row: state.active.row + rowDelta,
            column: state.active.column + columnDelta
        };
        if (!has$3(nextActive, state.model.data)) {
            return __assign(__assign({}, state), { mode: "view" });
        }
        return __assign(__assign({}, state), { active: nextActive, selected: create(nextActive, nextActive), mode: "view" });
    };
};
var keyDownHandlers = {
    ArrowUp: go(-1, 0),
    ArrowDown: go(+1, 0),
    ArrowLeft: go(0, -1),
    ArrowRight: go(0, +1),
    Tab: go(0, +1),
    Enter: edit,
    Backspace: clear,
    Delete: clear,
    Escape: blur
};
var editKeyDownHandlers = {
    Escape: view,
    Tab: keyDownHandlers.Tab,
    Enter: keyDownHandlers.ArrowDown
};
var editShiftKeyDownHandlers = {
    Tab: go(0, -1)
};
var shiftKeyDownHandlers = {
    ArrowUp: function (state) { return (__assign(__assign({}, state), { selected: modifyEdge(state.selected, state.active, state.model.data, Direction.Top) })); },
    ArrowDown: function (state) { return (__assign(__assign({}, state), { selected: modifyEdge(state.selected, state.active, state.model.data, Direction.Bottom) })); },
    ArrowLeft: function (state) { return (__assign(__assign({}, state), { selected: modifyEdge(state.selected, state.active, state.model.data, Direction.Left) })); },
    ArrowRight: function (state) { return (__assign(__assign({}, state), { selected: modifyEdge(state.selected, state.active, state.model.data, Direction.Right) })); },
    Tab: go(0, -1)
};
var shiftMetaKeyDownHandlers = {};
var metaKeyDownHandlers = {};
function getKeyDownHandler(state, event) {
    var key = event.key;
    var handlers;
    // Order matters
    if (state.mode === "edit") {
        if (event.shiftKey) {
            handlers = editShiftKeyDownHandlers;
        }
        else {
            handlers = editKeyDownHandlers;
        }
    }
    else if (event.shiftKey && event.metaKey) {
        handlers = shiftMetaKeyDownHandlers;
    }
    else if (event.shiftKey) {
        handlers = shiftKeyDownHandlers;
    }
    else if (event.metaKey) {
        handlers = metaKeyDownHandlers;
    }
    else {
        handlers = keyDownHandlers;
    }
    return handlers[key];
}
/** Returns whether the reducer has a handler for the given keydown event */
function hasKeyDownHandler(state, event) {
    return getKeyDownHandler(state, event) !== undefined;
}
/** Returns whether the active cell is read only */
function isActiveReadOnly(state) {
    var activeCell = getActive(state);
    return Boolean(activeCell === null || activeCell === void 0 ? void 0 : activeCell.readOnly);
}
/** Gets active cell from given state */
function getActive(state) {
    var activeCell = state.active && get$1(state.active, state.model.data);
    return activeCell || null;
}

var context = createContext([INITIAL_STATE, function () { }]);

function useDispatch() {
    return useContextSelector(context, function (_a) {
        var _b = __read(_a, 2); _b[0]; var dispatch = _b[1];
        return dispatch;
    });
}

function useSelector(selector) {
    return useContextSelector(context, function (_a) {
        var _b = __read(_a, 1), state = _b[0];
        return selector(state);
    });
}

var CornerIndicator = function (_a) {
    var selected = _a.selected, onSelect = _a.onSelect;
    var handleClick = React.useCallback(function () {
        onSelect();
    }, [onSelect]);
    return (React.createElement("th", { className: classNames("Spreadsheet__header", {
            "Spreadsheet__header--selected": selected
        }), onClick: handleClick, tabIndex: 0 }));
};
var enhance$3 = function (CornerIndicatorComponent) {
    return function CornerIndicatorWrapper(props) {
        var dispatch = useDispatch();
        var selectEntireTable$1 = React.useCallback(function () { return dispatch(selectEntireTable()); }, [dispatch]);
        var selected = useSelector(function (state) {
            return isEntireTable(state.selected);
        });
        return (React.createElement(CornerIndicatorComponent, __assign({}, props, { selected: selected, onSelect: selectEntireTable$1 })));
    };
};

var ColumnIndicator = function (_a) {
    var column = _a.column, label = _a.label, selected = _a.selected, onSelect = _a.onSelect;
    var handleClick = React.useCallback(function (event) {
        onSelect(column, event.shiftKey);
    }, [onSelect, column]);
    return (React.createElement("th", { className: classNames("Spreadsheet__header", {
            "Spreadsheet__header--selected": selected
        }), onClick: handleClick, tabIndex: 0 }, label !== undefined ? label : columnIndexToLabel(column)));
};
var enhance$2 = function (ColumnIndicatorComponent) {
    return function ColumnIndicatorWrapper(props) {
        var dispatch = useDispatch();
        var selectEntireColumn$1 = React.useCallback(function (column, extend) {
            return dispatch(selectEntireColumn(column, extend));
        }, [dispatch]);
        var selected = useSelector(function (state) {
            return hasEntireColumn(state.selected, props.column) ||
                isEntireTable(state.selected);
        });
        return (React.createElement(ColumnIndicatorComponent, __assign({}, props, { selected: selected, onSelect: selectEntireColumn$1 })));
    };
};
function columnIndexToLabel(column) {
    var label = "";
    var index = column;
    while (index >= 0) {
        label = String.fromCharCode(65 + (index % 26)) + label;
        index = Math.floor(index / 26) - 1;
    }
    return label;
}

var RowIndicator = function (_a) {
    var row = _a.row, label = _a.label, selected = _a.selected, onSelect = _a.onSelect;
    var handleClick = React.useCallback(function (event) {
        onSelect(row, event.shiftKey);
    }, [onSelect, row]);
    return (React.createElement("th", { className: classNames("Spreadsheet__header", {
            "Spreadsheet__header--selected": selected
        }), onClick: handleClick, tabIndex: 0 }, label !== undefined ? label : row + 1));
};
var enhance$1 = function (RowIndicatorComponent) {
    return function RowIndicatorWrapper(props) {
        var dispatch = useDispatch();
        var selected = useSelector(function (state) {
            return hasEntireRow(state.selected, props.row) ||
                isEntireTable(state.selected);
        });
        var selectEntireRow$1 = React.useCallback(function (row, extend) {
            return dispatch(selectEntireRow(row, extend));
        }, [dispatch]);
        return (React.createElement(RowIndicatorComponent, __assign({}, props, { selected: selected, onSelect: selectEntireRow$1 })));
    };
};

var Cell = function (_a) {
    var row = _a.row, column = _a.column, DataViewer = _a.DataViewer, selected = _a.selected, active = _a.active, dragging = _a.dragging, mode = _a.mode, data = _a.data, evaluatedData = _a.evaluatedData, select = _a.select, activate = _a.activate, setCellDimensions = _a.setCellDimensions, setCellData = _a.setCellData;
    var rootRef = React.useRef(null);
    var point = React.useMemo(function () { return ({
        row: row,
        column: column
    }); }, [row, column]);
    var handleMouseDown = React.useCallback(function (event) {
        if (mode === "view") {
            setCellDimensions(point, getOffsetRect(event.currentTarget));
            if (event.shiftKey) {
                select(point);
            }
            else {
                activate(point);
            }
        }
    }, [mode, setCellDimensions, point, select, activate]);
    var handleMouseOver = React.useCallback(function (event) {
        if (dragging) {
            setCellDimensions(point, getOffsetRect(event.currentTarget));
            select(point);
        }
    }, [setCellDimensions, select, dragging, point]);
    React.useEffect(function () {
        var root = rootRef.current;
        if (selected && root) {
            setCellDimensions(point, getOffsetRect(root));
        }
        if (root && active && mode === "view") {
            root.focus();
        }
    }, [setCellDimensions, selected, active, mode, point, data]);
    if (data && data.DataViewer) {
        // @ts-ignore
        DataViewer = data.DataViewer;
    }
    return (React.createElement("td", { ref: rootRef, className: classNames("Spreadsheet__cell", data === null || data === void 0 ? void 0 : data.className, {
            "Spreadsheet__cell--readonly": data === null || data === void 0 ? void 0 : data.readOnly
        }), onMouseOver: handleMouseOver, onMouseDown: handleMouseDown, tabIndex: 0 },
        React.createElement(DataViewer, { row: row, column: column, cell: data, evaluatedCell: evaluatedData, setCellData: setCellData })));
};
var enhance = function (CellComponent) {
    return function CellWrapper(props) {
        var row = props.row, column = props.column;
        var dispatch = useDispatch();
        var setCellData$1 = React.useCallback(function (data) {
            return dispatch(setCellData({ column: column, row: row }, data));
        }, [dispatch, column, row]);
        var select$1 = React.useCallback(function (point) { return dispatch(select(point)); }, [dispatch]);
        var activate$1 = React.useCallback(function (point) { return dispatch(activate(point)); }, [dispatch]);
        var setCellDimensions$1 = React.useCallback(function (point, dimensions) {
            return dispatch(setCellDimensions(point, dimensions));
        }, [dispatch]);
        var active = useSelector(function (state) {
            return isActive(state.active, {
                row: row,
                column: column
            });
        });
        var mode = useSelector(function (state) { return (active ? state.mode : "view"); });
        var data = useSelector(function (state) {
            return get$1({ row: row, column: column }, state.model.data);
        });
        var evaluatedData = useSelector(function (state) {
            return get$1({ row: row, column: column }, state.model.evaluatedData);
        });
        var selected = useSelector(function (state) {
            return hasPoint(state.selected, state.model.data, { row: row, column: column });
        });
        var dragging = useSelector(function (state) { return state.dragging; });
        var copied = useSelector(function (state) {
            return has$1({ row: row, column: column }, state.copied);
        });
        return (React.createElement(CellComponent, __assign({}, props, { selected: selected, active: active, copied: copied, dragging: dragging, mode: mode, evaluatedData: evaluatedData, data: data, select: select$1, activate: activate$1, setCellDimensions: setCellDimensions$1, setCellData: setCellData$1 })));
    };
};

var TRUE_TEXT = "TRUE";
var FALSE_TEXT = "FALSE";
/** The default Spreadsheet DataViewer component */
var DataViewer = function (_a) {
    var _b;
    var cell = _a.cell, evaluatedCell = _a.evaluatedCell;
    var value = (_b = evaluatedCell === null || evaluatedCell === void 0 ? void 0 : evaluatedCell.value) !== null && _b !== void 0 ? _b : cell === null || cell === void 0 ? void 0 : cell.value;
    return typeof value === "boolean" ? (React.createElement("span", { className: "Spreadsheet__data-viewer Spreadsheet__data-viewer--boolean" }, convertBooleanToText(value))) : (React.createElement("span", { className: "Spreadsheet__data-viewer" }, value));
};
function convertBooleanToText(value) {
    return value ? TRUE_TEXT : FALSE_TEXT;
}

/** The default Spreadsheet DataEditor component */
var DataEditor = function (_a) {
    var _b;
    var onChange = _a.onChange, cell = _a.cell;
    var inputRef = React.useRef(null);
    var handleChange = React.useCallback(function (event) {
        onChange(__assign(__assign({}, cell), { value: event.target.value }));
    }, [onChange, cell]);
    React.useEffect(function () {
        if (inputRef.current) {
            moveCursorToEnd(inputRef.current);
        }
    }, [inputRef]);
    var value = (_b = cell === null || cell === void 0 ? void 0 : cell.value) !== null && _b !== void 0 ? _b : "";
    return (React.createElement("div", { className: "Spreadsheet__data-editor" },
        React.createElement("input", { ref: inputRef, type: "text", onChange: handleChange, value: value, autoFocus: true })));
};

var ActiveCell = function (props) {
    var rootRef = React.useRef(null);
    var dispatch = useDispatch();
    var setCellData$1 = React.useCallback(function (active, data) {
        return dispatch(setCellData(active, data));
    }, [dispatch]);
    var edit = React.useCallback(function () { return dispatch(edit$1()); }, [dispatch]);
    var commit = React.useCallback(function (changes) {
        return dispatch(commit$1(changes));
    }, [dispatch]);
    var view = React.useCallback(function () {
        dispatch(view$1());
    }, [dispatch]);
    var active = useSelector(function (state) { return state.active; });
    var mode = useSelector(function (state) { return state.mode; });
    var cell = useSelector(function (state) {
        return state.active ? get$1(state.active, state.model.data) : undefined;
    });
    var dimensions = useSelector(function (state) {
        return active
            ? getCellDimensions(active, state.rowDimensions, state.columnDimensions)
            : undefined;
    });
    var hidden = React.useMemo(function () { return !active || !dimensions; }, [active, dimensions]);
    var initialCellRef = React.useRef(undefined);
    var prevActiveRef = React.useRef(null);
    var prevCellRef = React.useRef(undefined);
    var handleChange = React.useCallback(function (cell) {
        if (!active) {
            return;
        }
        setCellData$1(active, cell);
    }, [setCellData$1, active]);
    React.useEffect(function () {
        var root = rootRef.current;
        if (!hidden && root) {
            root.focus();
        }
    }, [rootRef, hidden]);
    React.useEffect(function () {
        var prevActive = prevActiveRef.current;
        var prevCell = prevCellRef.current;
        prevActiveRef.current = active;
        prevCellRef.current = cell;
        if (!prevActive || !prevCell) {
            return;
        }
        // Commit
        var coordsChanged = (active === null || active === void 0 ? void 0 : active.row) !== prevActive.row || (active === null || active === void 0 ? void 0 : active.column) !== prevActive.column;
        var exitedEditMode = mode !== "edit";
        if (coordsChanged || exitedEditMode) {
            var initialCell = initialCellRef.current;
            if (prevCell !== initialCell) {
                commit([
                    {
                        prevCell: initialCell || null,
                        nextCell: prevCell
                    },
                ]);
            }
            else if (!coordsChanged && cell !== prevCell) {
                commit([
                    {
                        prevCell: prevCell,
                        nextCell: cell || null
                    },
                ]);
            }
            initialCellRef.current = cell;
        }
    });
    var DataEditor = (cell && cell.DataEditor) || props.DataEditor;
    var readOnly = cell && cell.readOnly;
    return hidden ? null : (React.createElement("div", { ref: rootRef, className: classNames("Spreadsheet__active-cell", "Spreadsheet__active-cell--".concat(mode)), style: dimensions, onClick: mode === "view" && !readOnly ? edit : undefined, tabIndex: 0 }, mode === "edit" && active && (React.createElement(DataEditor, { row: active.row, column: active.column, cell: cell, 
        // @ts-ignore
        onChange: handleChange, exitEditMode: view }))));
};

var FloatingRect = function (_a) {
    var _b;
    var dimensions = _a.dimensions, dragging = _a.dragging, hidden = _a.hidden, variant = _a.variant;
    var _c = dimensions || {}, width = _c.width, height = _c.height, top = _c.top, left = _c.left;
    return (React.createElement("div", { className: classNames("Spreadsheet__floating-rect", (_b = {},
            _b["Spreadsheet__floating-rect--".concat(variant)] = variant,
            _b["Spreadsheet__floating-rect--dragging"] = dragging,
            _b["Spreadsheet__floating-rect--hidden"] = hidden,
            _b)), style: { width: width, height: height, top: top, left: left } }));
};

var Selected = function () {
    var selected = useSelector(function (state) { return state.selected; });
    var dimensions = useSelector(function (state) {
        return selected &&
            getSelectedDimensions(state.rowDimensions, state.columnDimensions, state.model.data, state.selected);
    });
    var dragging = useSelector(function (state) { return state.dragging; });
    var hidden = useSelector(function (state) { return size$2(state.selected, state.model.data) < 2; });
    return (React.createElement(FloatingRect, { variant: "selected", dimensions: dimensions, dragging: dragging, hidden: hidden }));
};

var Copied = function () {
    var range = useSelector(function (state) {
        return getCopiedRange(state.copied, state.hasPasted);
    });
    var dimensions = useSelector(function (state) {
        return range &&
            getRangeDimensions(state.rowDimensions, state.columnDimensions, range);
    });
    var hidden = range === null;
    return (React.createElement(FloatingRect, { variant: "copied", dimensions: dimensions, hidden: hidden, dragging: false }));
};

function styleInject(css, ref) {
  if ( ref === void 0 ) ref = {};
  var insertAt = ref.insertAt;

  if (!css || typeof document === 'undefined') { return; }

  var head = document.head || document.getElementsByTagName('head')[0];
  var style = document.createElement('style');
  style.type = 'text/css';

  if (insertAt === 'top') {
    if (head.firstChild) {
      head.insertBefore(style, head.firstChild);
    } else {
      head.appendChild(style);
    }
  } else {
    head.appendChild(style);
  }

  if (style.styleSheet) {
    style.styleSheet.cssText = css;
  } else {
    style.appendChild(document.createTextNode(css));
  }
}

var css_248z = ".Spreadsheet {\n  --background-color: white;\n  --text-color: black;\n  --readonly-text-color: rgba(0, 0, 0, 0.4);\n  --outline-color: #4285f4;\n  --outline-background-color: rgba(160, 195, 255, 0.2);\n  --border-color: hsl(2deg, 0%, 91%);\n  --header-background-color: rgba(0, 0, 0, 0.04);\n  --elevation: 0 2px 5px rgba(0, 0, 0, 0.4);\n\n  position: relative;\n  overflow: visible;\n  background: var(--background-color);\n  color: var(--text-color);\n  display: inline-block;\n}\n\n.Spreadsheet--dark-mode {\n  --background-color: black;\n  --text-color: white;\n  --readonly-text-color: rgba(255, 255, 255, 0.4);\n  --header-background-color: rgba(255, 255, 255, 0.04);\n  --border-color: hsl(2deg, 0%, 19%);\n}\n\n.Spreadsheet__active-cell {\n  position: absolute;\n  border: 2px solid var(--outline-color);\n  box-sizing: border-box;\n}\n\n.Spreadsheet__active-cell--edit {\n  background: var(--background-color);\n  box-shadow: var(--elevation);\n}\n\n.Spreadsheet__table {\n  border-collapse: collapse;\n  table-layout: fixed;\n}\n\n.Spreadsheet__cell,\n.Spreadsheet__active-cell {\n  cursor: cell;\n}\n\n.Spreadsheet__cell {\n  outline: none;\n}\n\n.Spreadsheet__cell--readonly {\n  color: var(--readonly-text-color);\n}\n\n.Spreadsheet__cell,\n.Spreadsheet__header {\n  min-width: 6em;\n  min-height: 1.9em;\n  height: 1.9em;\n  max-height: 1.9em;\n  border: 1px solid var(--border-color);\n  overflow: hidden;\n  word-break: keep-all;\n  white-space: nowrap;\n  text-align: left;\n  box-sizing: border-box;\n  user-select: none;\n}\n\n.Spreadsheet__header {\n  background: var(--header-background-color);\n  color: var(--readonly-text-color);\n  text-align: center;\n  font: inherit;\n}\n\n.Spreadsheet__header--selected {\n  background: #5f6268;\n  color: #fff;\n}\n\n.Spreadsheet__header,\n.Spreadsheet__data-viewer,\n.Spreadsheet__data-editor input {\n  padding: 4px;\n  box-sizing: border-box;\n}\n\n.Spreadsheet__data-editor,\n.Spreadsheet__data-editor input {\n  width: 100%;\n  height: 100%;\n}\n\n.Spreadsheet__data-editor input {\n  font: inherit;\n  color: inherit;\n  background: none;\n  border: none;\n  outline: none;\n  margin: 0;\n}\n\n.Spreadsheet__data-viewer--boolean {\n  text-align: center;\n}\n\n.Spreadsheet__floating-rect {\n  position: absolute;\n  pointer-events: none;\n  box-sizing: border-box;\n}\n\n.Spreadsheet__floating-rect--hidden {\n  display: none;\n}\n\n.Spreadsheet__floating-rect--selected {\n  background: var(--outline-background-color);\n  border: 2px var(--outline-color) solid;\n}\n\n.Spreadsheet__floating-rect--dragging {\n  border: none;\n}\n\n.Spreadsheet__floating-rect--copied {\n  border: 2px var(--outline-color) dashed;\n}\n";
styleInject(css_248z);

/**
 * The Spreadsheet component
 */
var Spreadsheet = function (props) {
    var className = props.className, darkMode = props.darkMode, columnLabels = props.columnLabels, rowLabels = props.rowLabels, hideColumnIndicators = props.hideColumnIndicators, hideRowIndicators = props.hideRowIndicators, onKeyDown = props.onKeyDown, _a = props.Table, Table$1 = _a === void 0 ? Table : _a, _b = props.Row, Row$1 = _b === void 0 ? Row : _b, _c = props.HeaderRow, HeaderRow$1 = _c === void 0 ? HeaderRow : _c, _d = props.DataEditor, DataEditor$1 = _d === void 0 ? DataEditor : _d, _e = props.DataViewer, DataViewer$1 = _e === void 0 ? DataViewer : _e, _f = props.onChange, onChange = _f === void 0 ? function () { } : _f, _g = props.onModeChange, onModeChange = _g === void 0 ? function () { } : _g, _h = props.onSelect, onSelect = _h === void 0 ? function () { } : _h, _j = props.onActivate, onActivate = _j === void 0 ? function () { } : _j, _k = props.onBlur, onBlur = _k === void 0 ? function () { } : _k, _l = props.onCellCommit, onCellCommit = _l === void 0 ? function () { } : _l;
    var initialState = React.useMemo(function () {
        var model = new Model(props.data);
        return __assign(__assign({}, INITIAL_STATE), { model: model });
    }, [props.data]);
    var reducerElements = React.useReducer(reducer, initialState);
    var _m = __read(reducerElements, 2), state = _m[0], dispatch = _m[1];
    var size = React.useMemo(function () {
        return calculateSpreadsheetSize(state.model.data, rowLabels, columnLabels);
    }, [state.model.data, rowLabels, columnLabels]);
    var mode = state.mode;
    var rootRef = React.useRef(null);
    var prevStateRef = React.useRef(initialState);
    var copy$1 = React.useCallback(function () { return dispatch(copy()); }, [dispatch]);
    var cut$1 = React.useCallback(function () { return dispatch(cut()); }, [dispatch]);
    var paste$1 = React.useCallback(function (data) { return dispatch(paste(data)); }, [dispatch]);
    var onKeyDownAction = React.useCallback(function (event) { return dispatch(keyDown(event)); }, [dispatch]);
    var onKeyPress = React.useCallback(function (event) { return dispatch(keyPress(event)); }, [dispatch]);
    var onDragStart = React.useCallback(function () { return dispatch(dragStart()); }, [dispatch]);
    var onDragEnd = React.useCallback(function () { return dispatch(dragEnd()); }, [dispatch]);
    var setData$1 = React.useCallback(function (data) { return dispatch(setData(data)); }, [dispatch]);
    var blur = React.useCallback(function () { return dispatch(blur$1()); }, [dispatch]);
    React.useEffect(function () {
        var e_1, _a;
        var prevState = prevStateRef.current;
        if (state.lastCommit && state.lastCommit !== prevState.lastCommit) {
            try {
                for (var _b = __values(state.lastCommit), _c = _b.next(); !_c.done; _c = _b.next()) {
                    var change = _c.value;
                    onCellCommit(change.prevCell, change.nextCell, state.lastChanged);
                }
            }
            catch (e_1_1) { e_1 = { error: e_1_1 }; }
            finally {
                try {
                    if (_c && !_c.done && (_a = _b["return"])) _a.call(_b);
                }
                finally { if (e_1) throw e_1.error; }
            }
        }
        if (state.model.data !== prevState.model.data) {
            // Call on change only if the data change internal
            if (state.model.data !== props.data) {
                onChange(state.model.data);
            }
        }
        if (state.selected !== prevState.selected) {
            var points = getPoints(state.selected, state.model.data);
            onSelect(points);
        }
        if (state.mode !== prevState.mode) {
            onModeChange(state.mode);
        }
        if (state.active !== prevState.active) {
            if (state.active) {
                onActivate(state.active);
            }
            else {
                var root = rootRef.current;
                if (root && isFocusedWithin(root) && document.activeElement) {
                    document.activeElement.blur();
                }
                onBlur();
            }
        }
        prevStateRef.current = state;
    }, [
        props.data,
        state,
        onActivate,
        onBlur,
        onCellCommit,
        onChange,
        onModeChange,
        onSelect,
        rowLabels,
        columnLabels,
    ]);
    React.useEffect(function () {
        var prevState = prevStateRef.current;
        if (props.data !== prevState.model.data) {
            setData$1(props.data);
        }
    }, [props.data, setData$1]);
    var clip = React.useCallback(function (event) {
        var model = state.model, selected = state.selected;
        var data = model.data;
        var selectedData = getSelectionFromMatrix(selected, data);
        var csv = getCSV(selectedData);
        writeTextToClipboard(event, csv);
    }, [state]);
    var handleCut = React.useCallback(function (event) {
        if (shouldHandleClipboardEvent(rootRef.current, mode)) {
            event.preventDefault();
            event.stopPropagation();
            clip(event);
            cut$1();
        }
    }, [mode, clip, cut$1]);
    var handleCopy = React.useCallback(function (event) {
        if (shouldHandleClipboardEvent(rootRef.current, mode)) {
            event.preventDefault();
            event.stopPropagation();
            clip(event);
            copy$1();
        }
    }, [mode, clip, copy$1]);
    var handlePaste = React.useCallback(function (event) {
        if (shouldHandleClipboardEvent(rootRef.current, mode)) {
            event.preventDefault();
            event.stopPropagation();
            if (event.clipboardData) {
                var text = readTextFromClipboard(event);
                paste$1(text);
            }
        }
    }, [mode, paste$1]);
    var handleKeyDown = React.useCallback(function (event) {
        event.persist();
        if (onKeyDown) {
            onKeyDown(event);
        }
        // Do not use event in case preventDefault() was called inside onKeyDown
        if (!event.defaultPrevented) {
            // Only disable default behavior if an handler exist
            if (hasKeyDownHandler(state, event)) {
                event.nativeEvent.preventDefault();
            }
            onKeyDownAction(event);
        }
    }, [state, onKeyDown, onKeyDownAction]);
    var handleMouseUp = React.useCallback(function () {
        onDragEnd();
        document.removeEventListener("mouseup", handleMouseUp);
    }, [onDragEnd]);
    var handleMouseMove = React.useCallback(function (event) {
        if (!state.dragging && event.buttons === 1) {
            onDragStart();
            document.addEventListener("mouseup", handleMouseUp);
        }
    }, [state, onDragStart, handleMouseUp]);
    var handleBlur = React.useCallback(function (event) {
        /**
         * Focus left self, Not triggered when swapping focus between children
         * @see https://reactjs.org/docs/events.html#detecting-focus-entering-and-leaving
         */
        if (!event.currentTarget.contains(event.relatedTarget)) {
            blur();
        }
    }, [blur]);
    var Cell$1 = React.useMemo(function () {
        // @ts-ignore
        return enhance(props.Cell || Cell);
    }, [props.Cell]);
    var CornerIndicator$1 = React.useMemo(function () {
        return enhance$3(props.CornerIndicator || CornerIndicator);
    }, [props.CornerIndicator]);
    var RowIndicator$1 = React.useMemo(function () { return enhance$1(props.RowIndicator || RowIndicator); }, [props.RowIndicator]);
    var ColumnIndicator$1 = React.useMemo(function () {
        return enhance$2(props.ColumnIndicator || ColumnIndicator);
    }, [props.ColumnIndicator]);
    React.useEffect(function () {
        document.addEventListener("cut", handleCut);
        document.addEventListener("copy", handleCopy);
        document.addEventListener("paste", handlePaste);
        return function () {
            document.removeEventListener("cut", handleCut);
            document.removeEventListener("copy", handleCopy);
            document.removeEventListener("paste", handlePaste);
        };
    }, [handleCut, handleCopy, handlePaste]);
    var tableNode = React.useMemo(function () { return (React.createElement(Table$1, { columns: size.columns, hideColumnIndicators: hideColumnIndicators },
        React.createElement(HeaderRow$1, null,
            !hideRowIndicators && !hideColumnIndicators && React.createElement(CornerIndicator$1, null),
            !hideColumnIndicators &&
                range(size.columns).map(function (columnNumber) {
                    return columnLabels ? (React.createElement(ColumnIndicator$1, { key: columnNumber, column: columnNumber, label: columnNumber in columnLabels
                            ? columnLabels[columnNumber]
                            : null })) : (React.createElement(ColumnIndicator$1, { key: columnNumber, column: columnNumber }));
                })),
        range(size.rows).map(function (rowNumber) { return (React.createElement(Row$1, { key: rowNumber, row: rowNumber },
            !hideRowIndicators &&
                (rowLabels ? (React.createElement(RowIndicator$1, { key: rowNumber, row: rowNumber, label: rowNumber in rowLabels ? rowLabels[rowNumber] : null })) : (React.createElement(RowIndicator$1, { key: rowNumber, row: rowNumber }))),
            range(size.columns).map(function (columnNumber) { return (React.createElement(Cell$1, { key: columnNumber, row: rowNumber, column: columnNumber, 
                // @ts-ignore
                DataViewer: DataViewer$1 })); }))); }))); }, [
        Table$1,
        size.rows,
        size.columns,
        hideColumnIndicators,
        Row$1,
        HeaderRow$1,
        hideRowIndicators,
        CornerIndicator$1,
        columnLabels,
        ColumnIndicator$1,
        rowLabels,
        RowIndicator$1,
        Cell$1,
        DataViewer$1,
    ]);
    var activeCellNode = React.useMemo(function () { return (React.createElement(ActiveCell
    // @ts-ignore
    , { 
        // @ts-ignore
        DataEditor: DataEditor$1 })); }, [DataEditor$1]);
    var rootNode = React.useMemo(function () { return (React.createElement("div", { ref: rootRef, className: classNames("Spreadsheet", className, {
            "Spreadsheet--dark-mode": darkMode
        }), onKeyPress: onKeyPress, onKeyDown: handleKeyDown, onMouseMove: handleMouseMove, onBlur: handleBlur },
        tableNode,
        activeCellNode,
        React.createElement(Selected, null),
        React.createElement(Copied, null))); }, [
        className,
        darkMode,
        onKeyPress,
        handleKeyDown,
        handleMouseMove,
        handleBlur,
        tableNode,
        activeCellNode,
    ]);
    return (React.createElement(context.Provider, { value: reducerElements }, rootNode));
};

export { DataEditor, DataViewer, Spreadsheet, createEmpty as createEmptyMatrix, Spreadsheet as default };
//# sourceMappingURL=index.js.map
